"""Before Data"""

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from typing import List, Tuple

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray
from pyarrow import Table

from ._epochs import BeForEpochs

ENC = "utf-8"


@dataclass
class BeForRecord:
    """Data Structure for handling behavioural force measurements

    Args
    ----
    dat: Pandas Dataframe
        data
    sample_rate: float
        the sampling rate of the force measurements
    columns: list of strings
        the columns in 'dat' that comprise the force measurements
    sessions: list of integer
        sample numbers at which a new recording session starts, if the exists
    time_column :
            str = ""
    meta: dictionary
        any kind of meta data
    """

    dat: pd.DataFrame
    sampling_rate: float
    columns: List[str] = field(default_factory=list[str])
    sessions: List[int] = field(default_factory=list[int])
    time_column: str = ""
    meta: dict = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.dat, pd.DataFrame):
            raise TypeError(f"must be pandas.DataFrame, not {type(self.dat)}")

        if isinstance(self.columns, str):
            self.columns = [self.columns]
        elif len(self.columns) == 0:
            # make col forces
            self.columns = self.dat.columns.values.tolist()

        if isinstance(self.sessions, int):
            self.sessions = [self.sessions]
        elif len(self.sessions) == 0:
            self.sessions.append(0)
        if len(self.time_column) > 0 and self.time_column not in self.dat:
            raise ValueError(f"Time column {self.time_column} not found in DataFrame")

    def __repr__(self):
        rtn = "BeForRecord"
        rtn += f"\n  sampling_rate: {self.sampling_rate}"
        rtn += f", n sessions: {self.n_sessions()}"
        rtn += f"\n  columns: {self.columns}".replace("[", "").replace("]", "")
        if len(self.time_column) >= 0:
            rtn += f"\n  time_column: {self.time_column}"
        rtn += "\n  metadata"
        for k, v in self.meta.items():
            rtn += f"\n  - {k}: {v}".rstrip()
        rtn += "\n" + str(self.dat)
        return rtn

    def add_session(self, dat: pd.DataFrame):
        """Adds data (dataframe) as a new recording

        Dataframe has to have the same columns as the already existing data
        """
        nbefore = self.dat.shape[0]
        self.dat = pd.concat([self.dat, dat], ignore_index=True)
        self.sessions.append(nbefore)

    def n_samples(self) -> int:
        """Number of sample in all sessions"""

        return self.dat.shape[0]

    def n_forces(self) -> int:
        """Number of force columns"""
        return len(self.columns)

    def n_sessions(self) -> int:
        """Number of recoding sessions"""
        return len(self.sessions)

    def time_stamps(self) -> NDArray:
        """The time stamps (numpy array)

        Creates time stamps, of they are not define in the data
        """
        if len(self.time_column) > 0:
            return self.dat.loc[:, self.time_column].to_numpy()
        else:
            step = 1000.0 / self.sampling_rate
            final_time = self.dat.shape[0] * step
            return np.arange(0, final_time, step)

    def session_samples(self, session: int) -> Tuple[int, int]:
        """tuple represents the range of the samples (from, to) of this sessions"""
        f = self.sessions[session]
        try:
            t = self.sessions[session + 1]
        except IndexError:
            t = self.dat.shape[0]
        return f, t - 1

    def get_data(
        self, columns: str | List[str] | None = None, session: int | None = None
    ) -> pd.DataFrame | pd.Series:
        """Returns data of a particular column and/or a particular session"""
        if columns is None:
            columns = self.dat.columns.values.tolist()

        if session is None:
            return self.dat.loc[:, columns]  # type: ignore
        else:
            f, t = self.session_samples(session)
            return self.dat.loc[f:t, columns]  # type: ignore

    def get_forces(self, session: int | None = None) -> pd.DataFrame | pd.Series:
        """Returns force data of a particular session"""
        return self.get_data(self.columns, session)

    def add_column(
        self, name: str, data: List | pd.Series, is_force_column: bool = True
    ):
        """Add data column (in place).

        Parameters
        ----------
        name : str
            columns name
        data : List or Pandas Series
            column data, that has to have the correct length
        is_force_column : bool, optional
            set this to False, if the added data do not comprise force
            measurements (default=true)
        """

        self.dat[name] = data
        if is_force_column:
            self.columns.append(name)

    def drop_column(self, name: str):
        """Drop a column for the data (in place)"""
        self.dat = self.dat.drop(name, axis=1)
        try:
            self.columns.remove(name)
        except ValueError:
            pass
        if self.time_column == name:
            self.time_column = ""

    def find_samples_by_time(self, times: ArrayLike) -> NDArray:
        """returns sample index (i) of the closes time in the BeForRecord.
        Takes the next larger element, if the exact time could not be found.

        .. math:: \\text{time_stamps}[i-1] <= t < \\text{time_stamps}[i]

        Parameters
        ----------
        timeline : ArrayLike
            the sorted array of time stamps

        """
        return np.searchsorted(self.time_stamps(), np.atleast_1d(times), "right")

    def extract_epochs(
        self,
        column: str,
        zero_samples: List[int] | NDArray[np.int_],
        n_samples: int,
        n_samples_before: int = 0,
        design: pd.DataFrame = pd.DataFrame(),
    ) -> BeForEpochs:
        """extracts epochs from BeForRecord

        Parameters
        ----------
        column: str
            name of column containing the force data to be used
        zero_samples: List[int]
            zero sample that define the epochs
        n_samples: int
            number of samples to be extract (from zero sample on)
        n_samples_before: int, optional
            number of samples to be extracted before the zero sample (default=0)

        design: pd.DataFrame, optional
            design information

        Notes
        -----
        use `find_times` to detect zero samples with time-based

        """

        fd = self.dat.loc[:, column]
        n = len(fd)  # samples for data
        n_epochs = len(zero_samples)
        n_col = n_samples_before + n_samples
        force_mtx = np.empty((n_epochs, n_col), dtype=np.float64)
        for r, zs in enumerate(zero_samples):
            f = zs - n_samples_before
            if f > 0 and f < n:
                t = zs + n_samples
                if t > n:
                    warnings.warn(
                        f"extract_force_epochs: last force epoch is incomplete, {t-n} samples missing.",
                        RuntimeWarning,
                    )
                    tmp = fd[f:]
                    force_mtx[r, : len(tmp)] = tmp
                    force_mtx[r, len(tmp) :] = 0
                else:
                    force_mtx[r, :] = fd[f:t]

        return BeForEpochs(
            force_mtx,
            sampling_rate=self.sampling_rate,
            design=design,
            zero_sample=n_samples_before,
        )

    def to_arrow(self) -> Table:
        """converts BeForRecord to ``pyarrow.Table``

        metadata of schema will be defined. Files can converted back to
        BeForRecord struct using ``BeForRecord.from_arrow()``

        Returns
        -------
        pyarrow.Table

        Examples
        --------
        >>> from pyarrow import feather
        >>> feather.write_feather(data.to_arrow(), "filename.feather",
                          compression="lz4", compression_level=6)
        """

        # Convert the DataFrame to a PyArrow table
        table = Table.from_pandas(self.dat, preserve_index=False)

        # Add metadata to the schema (serialize sampling_rate, timestamp, trigger, and meta)
        schema_metadata = {
            "sampling_rate": str(self.sampling_rate),
            "columns": ",".join(self.columns),
            "time_column": self.time_column,
            "sessions": ",".join([str(x) for x in self.sessions]),
        }
        schema_metadata.update(self.meta)
        return table.replace_schema_metadata(schema_metadata)

    @staticmethod
    def from_arrow(
        tbl: Table,
        sampling_rate: float | None = None,
        columns: str | List[str] | None = None,
        sessions: List[int] | None = None,
        time_column: str | None = None,
        meta: dict | None = None,
    ) -> BeForRecord:
        """Creates BeForRecord struct from `pyarrow.Table`

        Parameters
        ----------
        tbl : pyarrow.Table

        Examples
        --------
        >>> from pyarrow import feather
        >>> dat = feather.read_table("my_force_data.feather")
        >>> dat = BeforeData.from_arrow(dat)

        """

        if not isinstance(tbl, Table):
            raise TypeError(f"must be pyarrow.Table, not {type(tbl)}")

        if isinstance(columns, str):
            columns = [columns]

        # search info in meta data
        file_meta = {}
        if tbl.schema.metadata is not None:
            for k, v in tbl.schema.metadata.items():
                if k == b"sampling_rate":
                    if sampling_rate is None:
                        sampling_rate = float(v)
                elif k == b"columns":
                    if columns is None:
                        columns = v.decode(ENC).split(",")
                elif k == b"time_column":
                    if time_column is None:
                        time_column = v.decode(ENC)
                elif k == b"sessions":
                    if sessions is None:
                        sessions = [int(x) for x in v.decode(ENC).split(",")]
                else:
                    file_meta[k.decode(ENC)] = v.decode(ENC).strip()

        if sampling_rate is None:
            raise RuntimeError("No sampling rate defined!")
        if columns is None:
            columns = []
        if time_column is None:
            time_column = ""
        if sessions is None:
            sessions = []
        if isinstance(meta, dict):
            meta.update(file_meta)
        else:
            meta = file_meta

        return BeForRecord(
            dat=tbl.to_pandas(),
            sampling_rate=sampling_rate,
            columns=columns,  # type: ignore
            sessions=sessions,
            time_column=time_column,
            meta=meta,
        )



import gzip as _gzip
import lzma as _lzma
from io import StringIO as _StringIO
from pathlib import Path as _Path
from typing import List as _List

import pandas as _pd

from ._record import BeForRecord


def read_csv(
    file_path: str | _Path,
    columns: str | _List[str] | None = None,
    encoding: str = "utf-8",
    comment_char: str = "#",
):
    """Reads CSV file

    The function can handle comments as well as compressed CSVs, if they end
    with `.csv.xz` or `.csv.gz`

    Parameters
    ----------
    file_path : str | Path
        the path to the CSV file. If file end with `.csv.xz` or `.csv.gz`,
        decompression will be used
    columns : str | list[str], Optional
        the columns that should be read. If no columns are specified all columns
        are read
    encoding : str, optional
        file encoding, default="utf-8",

    comment_char : str, Optional
        line starting with character or string will be treated as comments and
        returned as a list of strings.

    Returns
    -------
    pandas.DataFrame and list[str] with all comments
    """

    p = _Path(file_path)
    if p.suffix.endswith("xz"):
        fl = _lzma.open(p, "rt", encoding=encoding)
    elif p.suffix.endswith("gz"):
        fl = _gzip.open(file_path, "rt", encoding=encoding)
    else:
        fl = open(file_path, "r", encoding=encoding)

    csv_str = ""
    comments = []
    for l in fl.readlines():
        if l.startswith(comment_char):
            comments.append(l)
        else:
            csv_str += l
    fl.close()

    df = _pd.read_csv(_StringIO(csv_str))
    if isinstance(columns, str):
        columns = [columns]
    if isinstance(columns, list):
        df = df.loc[:, columns]

    return df, comments


def read_csv_as_befordata(
    file_path: str | _Path,
    sampling_rate: float,
    columns: str | _List[str] | None = None,
    sessions: _List[int] | None = None,
    time_column: str | None = None,
    meta: dict | None = None,
    encoding: str = "utf-8",
    comment_char: str = "#",
):
    """Read CSV file as befordata

    Parameters
    ----------
    file_path: _description_
    sampling_rate: _description_
    columns: _description_. Defaults to None.
    sessions: _description_. Defaults to None.
    time_column: _description_. Defaults to None.
    meta: _description_. Defaults to None.
    encoding: _description_. Defaults to "utf-8".
    comment_char: _description_. Defaults to "#".

    Returns
    -------
    BeForRecord

    """

    df, _ = read_csv(file_path=file_path, encoding=encoding, comment_char=comment_char)
    if columns is None:
        columns = []
    elif isinstance(columns, str):
        columns = [columns]

    if time_column is None:
        time_column = ""
    if sessions is None:
        sessions = []
    if not isinstance(meta, dict):
        meta = {}

    return BeForRecord(
        dat=df,
        sampling_rate=sampling_rate,
        columns=columns,
        sessions=sessions,
        time_column=time_column,
        meta=meta,
    )



from copy import deepcopy
from typing import List

import numpy as np
import pandas as pd
from scipy import signal

from .._record import BeForRecord


def detect_sessions(
    data: BeForRecord, time_gap: float, time_column: str | None = None
) -> BeForRecord:
    """Detect recording sessions in the BeForRecord based on time gaps

    Parameters
    ----------
    data : BeForRecord
        the data
    time_gap : float
        smallest time gap that should be considered as pause of the recording
        and the start of a new session
    time_column : str | None, optional
        name of column that represents the time

    Returns
    -------
    BeForRecord
    """

    if time_column is None:
        time_column = data.time_column
    sessions = [0]
    breaks = np.flatnonzero(np.diff(data.dat[time_column]) >= time_gap) + 1
    sessions.extend(breaks.tolist())
    return BeForRecord(
        data.dat,
        sampling_rate=data.sampling_rate,
        columns=data.columns,
        sessions=sessions,
        time_column=time_column,
        meta=data.meta,
    )


def _butter_lowpass_filter(
    data: pd.Series, order: int, cutoff: float, sampling_rate: float, btype: str
):
    b, a = signal.butter(order, cutoff, fs=sampling_rate, btype=btype, analog=False)
    # filter shifted data (first sample = 0)
    y = signal.filtfilt(b, a, data - data.iat[0]) + data.iat[0]
    return y


def butter_filter(
    d: BeForRecord,
    order: int,
    cutoff: float,
    btype: str = "lowpass",
    columns: str | List[str] | None = None,
) -> BeForRecord:
    """Lowpass Butterworth filter of BeforRecord

    temporarily shifted data (first sample = 0) will be used for the filtering

    Parameters
    ----------
    d : BeForRecord
        the data
    order : int
        order of the filter.
    cutoff : float
        cutoff frequency
    btype : {‘lowpass’, ‘highpass’, ‘bandpass’, ‘bandstop’}, optional
        type of filter, by default "lowpass"
    columns : str | List[str] | None, optional
        column(s) of data the should be filtered. By default None; in this case,
        all data columns will be filtered

    Returns
    -------
    BeForRecord with filtered data

    Notes
    -----
    see documentation of `scipy.signal.butter` for information about the filtering

    """

    if columns is None:
        columns = d.columns
    elif not isinstance(columns, List):
        columns = [columns]

    df = d.dat.copy()
    for s in range(d.n_sessions()):
        f, t = d.session_samples(s)
        for c in columns:  # type: ignore
            df.loc[f:t, c] = _butter_lowpass_filter(
                data=df.loc[f:t, c],
                cutoff=cutoff,
                sampling_rate=d.sampling_rate,
                order=order,
                btype=btype,
            )
    meta = deepcopy(d.meta)
    meta["cutoff_freq"] = cutoff
    meta["butterworth_order"] = order
    return BeForRecord(
        df,
        sampling_rate=d.sampling_rate,
        columns=d.columns,
        sessions=d.sessions,
        time_column=d.time_column,
        meta=meta,
    )



"""Epochs Data Structure"""

from dataclasses import dataclass, field
from typing import Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray


@dataclass
class BeForEpochs:
    """Behavioural force data organized epoch-wis

    Args
    ----
    dat: : 2d numpy array
        data. Each row of the 2D numpy array represents one epoch. Thus, the number
        of rows equals the number of epochs and number of columns equals the number
        of samples in each epoch.

    sample_rate: float
        sampling rate of the force measurements

    design : pd.DataFrame
        design data frame

    baseline : numpy array
        baseline for each epoch at `zero_sample`

    zero_sample : int, optional
        sample index that represents the time 0

    """

    dat: NDArray[np.floating]
    sampling_rate: float
    design: pd.DataFrame = field(default_factory=pd.DataFrame())  # type: ignore
    baseline: NDArray[np.float64] = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )
    zero_sample: int = 0

    def __post_init__(self):
        self.dat = np.atleast_2d(self.dat)
        if self.dat.ndim != 2:
            raise ValueError("Epoch data but be a 2D numpy array")

        ne = self.n_epochs()
        if self.design.shape[0] > 0 and self.design.shape[0] != ne:
            raise ValueError("Epoch data and design must have the same number of rows")

        self.baseline = np.atleast_1d(self.baseline)
        if self.baseline.ndim != 1 and len(self.baseline) != ne:
            raise ValueError(
                "baseline must be a 1D array. The number of elements must match the of epochs"
            )

    def __repr__(self):
        rtn = "BeForEpochs"
        rtn += f"\n  n epochs: {self.n_epochs()}"
        rtn += f", n_samples: {self.n_samples()}"
        rtn += f"\n  sampling_rate: {self.sampling_rate}"
        rtn += f", zero_sample: {self.zero_sample}"
        if len(self.design) == 0:
            rtn += "\n  design: None"
        else:
            rtn += f"\n  design: {list(self.design.columns)}".replace("[", "").replace(
                "]", ""
            )
        # rtn += "\n" + str(self.dat)
        return rtn

    def n_epochs(self) -> int:
        """number of epochs"""
        return self.dat.shape[0]

    def n_samples(self) -> int:
        """number of sample of one epoch"""
        return self.dat.shape[1]

    def adjust_baseline(self, reference_window: Tuple[int, int]):
        """Adjust the baseline of each epoch using the mean value of
        a defined range of sample (reference window)

        Parameters
        ----------
        reference_window : Tuple[int, int]
            sample range that is used for the baseline adjustment

        """

        if len(self.baseline) > 0:
            dat = self.dat + np.atleast_2d(self.baseline).T  # rest baseline
        else:
            dat = self.dat
        i = range(reference_window[0], reference_window[1])
        self.baseline = np.mean(dat[:, i], axis=1)
        self.dat = dat - np.atleast_2d(self.baseline).T

    def to_arrow(self):
        # FIXME
        pass
        pass
